require 'rubygems'
require 'fiber'
require 'uuid'
require 'bunny'
require 'net/http'

module Wire

  class Function
    def initialize(name, signature = nil, result_type = nil, native_type = nil)
      @name = name
      @signature = signature
      @result_type = result_type
      @native_type = native_type
    end

    # ignore this for any real service call, the http endpoint is evidently incapable of getting this from a header
    def service= (service)
      @service=service
    end

    # ignore this for any real service call, the http endpoint is evidently incapable of getting this from a header
    def version= (version)
      @version=version
    end

    def to_json
      "{'endpoint': '#{@name}'" + ((@signature == nil) ? '' : ", 'signature': #{@signature}") + ((@result_type == nil) ? '' : ", 'resultType': '#{@result_type}'") + ((@native_type == nil) ? '' : ", 'nativeType': '#{@native_type}'") + ((@service == nil) ? '' : ", 'service': '#{@service}'") + ((@version == nil) ? '' : ", 'version': '#{@version}'") + '}'
    end
  end

  class InvocationSignal
    def initialize(function, arguments, contexts)
      @function = function
      @arguments = arguments
      @contexts = contexts
    end

    def function
      @function
    end

    def contexts= (contexts)
      @contexts=contexts
    end

    def contexts
      @contexts
    end

    def to_json
      "{'address': #{@function.to_json}, 'body': #{@arguments.to_json}, 'contexts': #{@contexts.to_json}}"
    end
  end

  class MessageCallback
    def initialize (message_hash, message_id, callback, timeout_seconds)
      @semaphore = Mutex.new
      @callback = callback
      @result = nil

      Thread.new do
        sleep(timeout_seconds)
        message_hash.delete(message_id)
        @semaphore.synchronize do
          if @result.nil?
            callback.timeout
          end
        end
      end
    end

    def set_result (result)
      @semaphore.synchronize do
        @result = result
        @callback.complete(@result)
      end
    end
  end

  class Message

    @result = nil

    def result
      while @result == nil
        sleep(0.1)
      end

      @result
    end

    def complete (result)
      @result = result
    end

    def timeout
      raise Timeout::Error.new
    end
  end

  class RabbitMQTransport
    def initialize (host, port)
      @connection = Bunny.new("amqp://#{host}:#{port}")
      @connection.start

      @semaphore = Mutex.new

      @uuid = UUID.new
      @caller_id = @uuid.generate

      @topic_hash = {}
      @message_hash = {}
    end

    def initialize_service (service_name)
      @semaphore.synchronize do
        topic = @topic_hash[service_name]

        if topic == nil
          request_channel = @connection.create_channel
          topic = request_channel.topic("#{service_name}.request")

          response_queue_name = @uuid.generate
          response_channel = @connection.create_channel
          response_exchange = response_channel.topic("#{service_name}.response")
          response_queue = response_channel.queue("#{response_queue_name}", {:durable => false, :exclusive => true, :auto_delete => true})

          response_queue.bind(response_exchange, :routing_key => "#{@caller_id}").subscribe do |delivery_info, metadata, payload|
            if (message_callback = @message_hash[metadata[:correlation_id]]) != nil
              result_body = JSON.parse(payload)
              message_callback.set_result(result_body)
            end
          end
        end

        return topic
      end
    end

    def transmit_in_only(service_name, version, invocation_signal)
      initialize_service(service_name).publish(invocation_signal.to_json, :routing_key => "#{version}.2", :headers => {'x-opt-callerId' => "#{@caller_id}"}, :content_type => 'application/json', :message_id => "#{@uuid.generate}", :timestamp => Time.new.to_i)
    end

    def transmit(service_name, version, invocation_signal, timeout_seconds)
      message_id = "#{@uuid.generate}".freeze
      @message_hash[message_id] = MessageCallback.new(@message_hash, message_id, message = Message.new, timeout_seconds)
      initialize_service(service_name).publish(invocation_signal.to_json, :routing_key => "#{version}.2", :headers => {'x-opt-callerId' => "#{@caller_id}"}, :content_type => 'application/json', :message_id => "#{message_id}", :timestamp => Time.new.to_i)

      message
    end
  end

  class HttpTransport
    def initialize (host, port = nil)
      @type_signatures = {'Fixnum'.freeze => 'I', 'String'.freeze => 'G', 'Array'.freeze => '!Enumerable', 'Hash'.freeze => '!Dictionary'}

      @host = host
      if port != nil
        @host += ':' + port
      end
    end

    def transmit (service_name, version, invocation_signal, timeout_seconds)
      message = Message.new

      begin
        uri = URI("#{@host}/api/message")
        headers = {'Content-Type' => 'application/json'}
        http = Net::HTTP.new(uri.host, uri.port)
        http.read_timeout = timeout_seconds

        invocation_signal.function.service = service_name
        invocation_signal.function.version = version

        context_hash = {}
        invocation_signal.contexts.each { |context| context.each { |key, value| context_hash[key] = value } }
        invocation_signal.contexts = context_hash

        res = http.post(uri.path, invocation_signal.to_json, headers)

        case res
          when Net::HTTPSuccess
            response = JSON.parse res.body
            message.complete(response)
          else
            res.error!
        end
      rescue Timeout::Error
        message.timeout
      end

      message
    end

    # call a service method
    def call(service, method, args = nil, result_type = nil)
      function = Function.new(method, (args == nil) ? nil : get_signature(args), result_type)
      contexts = [
          {'auth'.freeze => {'user'.freeze => {'id'.freeze => '1'}, 'organization'.freeze => {'id'.freeze => '1'}, 'proxy'.freeze => {'id'.freeze => '-1'}}},
          {'core'.freeze => {'transaction'.freeze => {'id'.freeze => UUID.new.generate.to_s}, 'stream'.freeze => {'id'.freeze => UUID.new.generate.to_s}}}
      ]

      #  puts 'Calling ' + service + '.' + method #+ signature.to_s
      message = self.transmit(service, '1', InvocationSignal.new(function, args, contexts), 60)
      result = message.result
      return result['body']
    end

    # return array with the type of each argument
    def get_signature(args)
      arg_types=[]
      args.each do |arg|
        arg_types.push(get_type(arg))
      end
      return arg_types
    end

    def get_type (arg)
      return @type_signatures[arg.class.to_s]
    end
  end
end